"""Docker Compose YAML file generator.

This package provides utilities for programmatically creating and managing Docker Compose
configuration files through Python classes, simplifying the process of defining
and managing Docker Compose configurations.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Self

import yaml
from pydantic import BaseModel, ValidationError


class ServiceConfig(BaseModel):
    image: str = ""
    ports: list[str] | None = None
    environment: dict[str, str] | None = None
    volumes: list[str] | None = None
    networks: list[str] | None = None
    command: str | list[str] | None = None
    entrypoint: str | list[str] | None = None
    build: dict[str, Any] | str | None = None
    healthcheck: dict[str, Any] | None = None
    restart: str | None = None
    labels: dict[str, str] | list[str] | None = None
    depends_on: list[str] | dict[str, dict[str, str]] | None = None
    configs: list[dict[str, Any]] | None = None
    secrets: list[dict[str, Any]] | None = None
    deploy: dict[str, Any] | None = None
    resources: dict[str, Any] | None = None

    # Allow extra fields for flexibility and to support arbitrary Docker Compose extensions.
    model_config = {"extra": "allow"}


class VolumeConfig(BaseModel):
    """Configuration for a Docker Compose volume.

    Represents the configuration options for defining a volume in a Docker Compose file.
    """

    driver: str | None = None
    driver_opts: dict[str, Any] | None = None
    external: bool | dict[str, Any] | None = None
    labels: dict[str, str] | list[str] | None = None
    name: str | None = None

    # Allow extra fields for flexibility and to support arbitrary Docker Compose extensions.
    model_config = {"extra": "allow"}


class NetworkConfig(BaseModel):
    """Represents a network configuration for a Docker Compose file."""

    driver: str | None = None
    driver_opts: dict[str, Any] | None = None
    attachable: bool | None = None
    external: bool | dict[str, Any] | None = None
    labels: dict[str, str] | list[str] | None = None
    name: str | None = None

    # Allow extra fields for flexibility and to support arbitrary Docker Compose extensions.
    model_config = {"extra": "allow"}


class ConfigConfig(BaseModel):
    """Represents a config configuration for a Docker Compose file."""

    file: str | None = None
    external: bool | None = None
    name: str | None = None

    # Allow extra fields for flexibility and to support arbitrary Docker Compose extensions.
    model_config = {"extra": "allow"}


class SecretConfig(BaseModel):
    """Represents a secret configuration for a Docker Compose file."""

    file: str | None = None
    external: bool | None = None
    name: str | None = None

    # Allow extra fields for flexibility and to support arbitrary Docker Compose extensions.
    model_config = {"extra": "allow"}


if TYPE_CHECKING:
    from types import TracebackType


class DockerComposeManager:
    """A class to create and modify Docker Compose YAML files programmatically.

    Supports context manager usage for auto-saving.
    """

    def add_volume(self, name: str, config: VolumeConfig | dict[str, Any] | None = None) -> DockerComposeManager:
        """Add a top-level volume definition.

        Returns:
            DockerComposeManager: self (for chaining)
        """
        if "volumes" not in self._data or not isinstance(self._data["volumes"], dict):
            self._data["volumes"] = {}
        if config is None:
            self._data["volumes"][name] = {}
        else:
            if isinstance(config, dict):
                config = VolumeConfig(**config)
            self._data["volumes"][name] = config.model_dump(exclude_none=True)
        self._dirty = True
        return self

    def remove_volume(self, name: str) -> DockerComposeManager:
        """Remove a top-level volume definition.

        Returns:
            DockerComposeManager: self (for chaining)

        Raises:
            KeyError: If the volume does not exist.
        """
        if "volumes" not in self._data or name not in self._data["volumes"]:
            msg = f"Volume '{name}' not found."
            raise KeyError(msg)
        del self._data["volumes"][name]
        if not self._data["volumes"]:
            del self._data["volumes"]
        self._dirty = True
        return self

    def add_network(self, name: str, config: NetworkConfig | dict[str, Any] | None = None) -> DockerComposeManager:
        """Add a top-level network definition.

        Returns:
            DockerComposeManager: self (for chaining)
        """
        if "networks" not in self._data or not isinstance(self._data["networks"], dict):
            self._data["networks"] = {}
        if config is None:
            self._data["networks"][name] = {}
        else:
            if isinstance(config, dict):
                config = NetworkConfig(**config)
            self._data["networks"][name] = config.model_dump(exclude_none=True)
        self._dirty = True
        return self

    def remove_network(self, name: str) -> DockerComposeManager:
        """Remove a top-level network definition.

        Returns:
            DockerComposeManager: self (for chaining)

        Raises:
            KeyError: If the network does not exist.
        """
        if "networks" not in self._data or name not in self._data["networks"]:
            msg = f"Network '{name}' not found."
            raise KeyError(msg)
        del self._data["networks"][name]
        if not self._data["networks"]:
            del self._data["networks"]
        self._dirty = True
        return self

    def add_config(self, name: str, config: ConfigConfig | dict[str, Any] | None = None) -> DockerComposeManager:
        """Add a top-level config definition.

        Returns:
            DockerComposeManager: self (for chaining)
        """
        if "configs" not in self._data or not isinstance(self._data["configs"], dict):
            self._data["configs"] = {}
        if config is None:
            self._data["configs"][name] = {}
        else:
            if isinstance(config, dict):
                config = ConfigConfig(**config)
            self._data["configs"][name] = config.model_dump(exclude_none=True)
        self._dirty = True
        return self

    def remove_config(self, name: str) -> DockerComposeManager:
        """Remove a top-level config definition.

        Returns:
            DockerComposeManager: self (for chaining)

        Raises:
            KeyError: If the config does not exist.
        """
        if "configs" not in self._data or name not in self._data["configs"]:
            msg = f"Config '{name}' not found."
            raise KeyError(msg)
        del self._data["configs"][name]
        if not self._data["configs"]:
            del self._data["configs"]
        self._dirty = True
        return self

    def add_secret(self, name: str, config: SecretConfig | dict[str, Any] | None = None) -> DockerComposeManager:
        """Add a top-level secret definition.

        Returns:
            DockerComposeManager: self (for chaining)
        """
        if "secrets" not in self._data or not isinstance(self._data["secrets"], dict):
            self._data["secrets"] = {}
        if config is None:
            self._data["secrets"][name] = {}
        else:
            if isinstance(config, dict):
                config = SecretConfig(**config)
            self._data["secrets"][name] = config.model_dump(exclude_none=True)
        self._dirty = True
        return self

    def remove_secret(self, name: str) -> DockerComposeManager:
        """Remove a top-level secret definition.

        Returns:
            DockerComposeManager: self (for chaining)

        Raises:
            KeyError: If the secret does not exist.
        """
        if "secrets" not in self._data or name not in self._data["secrets"]:
            msg = f"Secret '{name}' not found."
            raise KeyError(msg)
        del self._data["secrets"][name]
        if not self._data["secrets"]:
            del self._data["secrets"]
        self._dirty = True
        return self

    def __init__(self, path: str, version: str = "3.8") -> None:
        """Initialize the manager with a YAML file path. Loads existing file or creates a new one."""
        self.path: str = path
        self.version: str = version
        self._data: dict[str, Any] = {}
        self._dirty: bool = False
        self._load()

    def _load(self) -> None:
        if Path(self.path).exists():
            with Path(self.path).open("r", encoding="utf-8") as f:
                self._data = yaml.safe_load(f) or {}
        if not self._data:
            self._data = {"version": self.version, "services": {}}
        if "services" not in self._data:
            self._data["services"] = {}
        if not isinstance(self._data["services"], dict):
            self._data["services"] = {}

    def create_service(
        self,
        name: str,
        *,
        config: ServiceConfig | dict[str, Any] | None = None,
        image: str = "",
        ports: list[str] | None = None,
        environment: dict[str, str] | None = None,
        volumes: list[str] | None = None,
        networks: list[str] | None = None,
        command: str | list[str] | None = None,
        entrypoint: str | list[str] | None = None,
        build: dict[str, Any] | str | None = None,
        healthcheck: dict[str, Any] | None = None,
        restart: str | None = None,
        labels: dict[str, str] | list[str] | None = None,
        depends_on: list[str] | dict[str, dict[str, str]] | None = None,
        configs: list[dict[str, Any]] | None = None,
        secrets: list[dict[str, Any]] | None = None,
        deploy: dict[str, Any] | None = None,
        resources: dict[str, Any] | None = None,
        **kwargs: object,
    ) -> DockerComposeManager:
        """Create a new service in the compose file.

        Returns:
            DockerComposeManager: self (for chaining)

        Raises:
            ValueError: If the service config is invalid.
        """
        services = self._data["services"]
        if config is not None:
            if isinstance(config, dict):
                config = ServiceConfig(**config)
            service = config.model_dump(exclude_none=True)
            service.update(kwargs)
        else:
            try:
                service = ServiceConfig(
                    image=image,
                    ports=ports,
                    environment=environment,
                    volumes=volumes,
                    networks=networks,
                    command=command,
                    entrypoint=entrypoint,
                    build=build,
                    healthcheck=healthcheck,
                    restart=restart,
                    labels=labels,
                    depends_on=depends_on,
                    configs=configs,
                    secrets=secrets,
                    deploy=deploy,
                    resources=resources,
                    **kwargs,
                ).model_dump(exclude_none=True)
            except ValidationError as e:
                msg = f"Invalid service config: {e}"
                raise ValueError(msg) from e
        services[name] = service
        self._dirty = True
        return self

    def modify_service(self, name: str, **kwargs: object) -> DockerComposeManager:
        """Modify an existing service. Raises KeyError if not found.

        Args:
            name (str): Name of the service to modify.
            **kwargs: Key-value pairs to update in the service configuration.

        Raises:
            KeyError: If the service with the given name does not exist.

        Returns:
            DockerComposeManager: self (for chaining)

        """
        services: dict[str, dict[str, Any]] = self._data["services"]
        if name not in services:
            msg: str = f"Service '{name}' not found."
            raise KeyError(msg)
        services[name].update(kwargs)
        self._dirty = True
        return self

    def remove_service(self, name: str) -> DockerComposeManager:
        """Remove a service from the compose file.

        Returns:
            DockerComposeManager: self (for chaining)

        Raises:
            KeyError: If the service does not exist.
        """
        services: dict[str, dict[str, Any]] = self._data["services"]
        if name not in services:
            msg: str = f"Service '{name}' not found."
            raise KeyError(msg)
        del services[name]
        self._dirty = True
        return self

    def save(self) -> None:
        """Save the current state to the YAML file."""
        with Path(self.path).open("w", encoding="utf-8") as f:
            yaml.dump(self._data, f, sort_keys=False, indent=2, default_flow_style=False)
        self._dirty = False

    def __enter__(self) -> Self:
        """Enter the context manager and return self.

        Returns:
            DockerComposeManager: The instance itself for context management.

        """
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the context manager and save the file if changes were made.

        Args:
            exc_type: The exception type if an exception was raised, None otherwise.
            exc_val: The exception value if an exception was raised, None otherwise.
            exc_tb: The traceback if an exception was raised, None otherwise.

        """
        if self._dirty:
            self.save()
